import gettext
import importlib.util
import os
import sys
import warnings

try:
    import PySimpleGUI as sg
    GUI_ENABLED = True
    del sg
except ImportError as e:
    if e.name != 'PySimpleGUI':
        # GUI was installed, but failed to load
        # due to tkinter missing or other dependencies.
        warnings.warn('Failed to load GUI: %s' % e.msg)
    GUI_ENABLED = False


def print_error(error_msg, message_queue=None):
    """Print error message to `stderr` and dublicate it to
    `message_queue` if it's been passed.

    Args:
        error_msg (str): Error message.
        message_queue (multiprocessing.Queue): Message queue.
    """
    print(error_msg, file=sys.stderr)
    if message_queue:
        # Send error message up to the message queue
        message_queue.put(('#Error', error_msg))


def data_path():
    """Returns data path."""
    if '_MEIPASS2' in os.environ:
        here = os.environ['_MEIPASS2']
    else:
        here = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(here, 'data')


def image_path(image_name):
    """Get package's data path.

    Args:
        image_name (str): Image basename.

    Returns:
        Image full path.
    """
    return os.path.join(data_path(), 'images', image_name)


def patch_tqdm(tqdm, message_queue):
    """Patch tqdm to make it dublicate progress to `message_queue`.

    Args:
        tqdm (tqdm.tqdm) Tqdm class to be patched.
        message_queue (multiprocessing.Queue): Message queue.
    """
    original_update = tqdm.update
    original_close = tqdm.close

    def new_update(tqdm_self, *args, **kwargs):
        original_update(tqdm_self, *args, **kwargs)
        message_queue.put(('#UpdateProgressBar', (tqdm_self.n, tqdm_self.total)))

    def new_close(tqdm_self, *args, **kwargs):
        message_queue.put(('#UpdateProgressBar', (tqdm_self.n, tqdm_self.total)))
        original_close(tqdm_self, *args, **kwargs)

    tqdm.update = new_update
    tqdm.close = new_close


def get_files_recursively(root_path, exts):
    """Recursively search for files with the extension
    specified in `exts`.

    Args:
        root_path (str): Root path.
        exts (str): Files extension.

    Returns:
        List of found files.
    """
    if type(exts) not in (list, tuple):
        exts = (exts,)
    result = []
    files = [os.path.join(dp, f) for dp, _, fn in os.walk(root_path) for f in fn]
    for f in files:
        if any(f.endswith(ext) for ext in exts):
            full_path_f = os.path.join(root_path, f)
            result.append(full_path_f)
    return result


def install_i18n():
    """Install i18n."""
    locale_dir = os.path.join(data_path(), 'locale')
    if not os.path.isdir(locale_dir):
        # find locale in module path
        qdc_converter_module = importlib.util.find_spec('qdc-converter')
        if qdc_converter_module:
            for path in qdc_converter_module.submodule_search_locations:
                locale_path = os.path.join(path, 'data', 'locale')
                if os.path.exists(locale_path):
                    locale_dir = locale_path

    gettext.install('messages', locale_dir)



import base64
import multiprocessing as mp
import os
import threading

import PySimpleGUI as sg

from .cli import run_cli
from .utils import get_files_recursively, image_path


def run_gui(qdc_folder_path, output_path, layer, validity_codes, quite, x_correction,
            y_correction, z_correction, csv_delimiter, csv_skip_headers, csv_yxz, multithreaded):
    """Run GUI with same passed arguments as CLI."""
    sg.theme('SandyBeach')

    # Work with local variables through `args`
    args = locals().copy()

    # Set icon
    with open(image_path('icon.png'), 'rb') as f:
        sg.set_global_icon(base64.b64encode(f.read()))

    # Window's text, labels
    app_description = _('QDC Converter.\n\nConverter of Garmin\'s QDC files into CSV or GRD.')
    window_title, window_description = app_description.split('\n\n')

    t = lambda title, tail='': title[:-1] + tail
    input_dir_title = t(_('Path to folder with QuickDraw Contours (QDC) inside.'), tail=':')
    output_file_title = t(_('Path to the result file (*.csv or *.grd).'), tail=':')
    layer_title = t(_('Data layer (0 - Raw user data, 1 - Recommended).'), tail=':')
    validity_codes_title = t(_('Write validity code instead of depth.'))
    multithreaded_title = t(_('Enable multithreading.'))
    csv_skip_headers_title = t(_('Do not write header.'))
    csv_yxz_title = t(_('Change column order from X,Y,Z to Y,X,Z.'),)
    window_description = t(window_description)
    window_title = t(window_title)

    # Window's layout
    layout = [
        [
            sg.Text(window_description, font='Any 15')
        ],
        [
            sg.Frame('', expand_x=True, layout=[
                [
                    sg.Text(input_dir_title),
                ],
                [
                    sg.Input(qdc_folder_path, key='@qdc_folder_path', expand_x=True),
                    sg.FolderBrowse(_('Browse')),
                ],
            ])
        ],
        [
            sg.Frame('', expand_x=True, layout=[
                [
                    sg.Text(output_file_title),
                ],
                [
                    sg.Input(output_path, key='@output_path', expand_x=True),
                    sg.FileSaveAs(_('Browse'), file_types=(('CSV Table', '*.csv'), ('ESRI ASCII grid', '*.grd')))
                ],
            ]),
        ],
        [
            sg.Frame(_('Options'), expand_x=True, layout=[
                [
                    sg.Text(layer_title),
                    sg.Combo(key='@layer', default_value=1 if layer is None else layer,
                             size=(3, 1), values=(0, 1, 2, 3, 4, 5), readonly=True),
                ],
                [
                    sg.Checkbox(validity_codes_title, key='@validity_codes', default=validity_codes),
                ],
                [
                    sg.Checkbox(multithreaded_title, key='@multithreaded', default=multithreaded),
                ],
            ]),
        ],
        [
            sg.Frame('CSV', expand_x=True, layout=[
                [
                    sg.Checkbox(csv_skip_headers_title, key='@csv_skip_headers', default=csv_skip_headers),
                ],
                [
                    sg.Checkbox(csv_yxz_title, key='@csv_yxz', default=csv_yxz),
                ],
            ]),
        ],
        [
            sg.ProgressBar(0, orientation='h', size=(1, 22), expand_x=True, key='-ProgressBar-', pad=(5, 10))
        ],
        [
            sg.Column([[
                sg.Button(_('Convert'), key='-Convert-'),
                sg.Button(_('Cancel'), key='-Cancel-', button_color=('white', 'orange3'), visible=False),
            ]], pad=0),
            sg.Button(_('Quit'), key='-Quit-', button_color=('white', 'firebrick3')),
        ],
    ]

    # Main window
    window = sg.Window(window_title, layout, auto_size_text=True, auto_size_buttons=False,
                       default_element_size=(20, 1), text_justification='right', font='Any 12')

    # Process
    converter_process = None
    converter_process_running = lambda: converter_process and converter_process.is_alive()

    # Window elements
    cancel_button = window['-Cancel-']
    convert_button = window['-Convert-']
    progress_bar = window['-ProgressBar-']

    # IPC message queue
    message_queue = mp.Queue()

    # Create bridge between MP message queue and SG message queue
    def message_queue_bridge():
        # Perform message resend from subprocess
        # message queue to parent GUI message queue.
        while True:
            key, value = message_queue.get()
            window.write_event_value(key, value)

    message_bridge_thread = threading.Thread(target=message_queue_bridge, daemon=True)
    message_bridge_thread.start()

    # Add message queue to args
    args['message_queue'] = message_queue

    def swap_buttons(state):
        if state:
            cancel_button.update(visible=True)
            convert_button.update(visible=False)
        else:
            cancel_button.update(visible=False)
            convert_button.update(visible=True)

    # Main event loop
    while True:
        event, values = window.read()

        if event in ('-Quit-', None):
            if converter_process_running():
                converter_process.terminate()
                converter_process.join()
            break

        elif event == '#Error':
            sg.PopupError(values['#Error'], title=_('Error'), font='Any 12')

        elif event == '#UpdateProgressBar':
            current_value, max_value = values['#UpdateProgressBar']
            progress_bar.update(current_count=current_value, max=max_value)

        elif event == '-Cancel-':
            if converter_process_running():
                converter_process.terminate()

        elif event == '-Convert-':
            # Update arguments
            for name, value in values.items():
                if name.startswith('@'):
                    args[name[1:]] = value

            # Validate input path exists
            if not os.path.isdir(args['qdc_folder_path']):
                sg.PopupError(_('Input path is not a folder!'), title=_('Error'), font='Any 12')
                continue

            # Validate input path contains *.qdc files
            qdc_files = get_files_recursively(args['qdc_folder_path'], '.qdc')
            if not qdc_files:
                sg.PopupError(_('No *.qdc files found inside input path!'), title=_('Error'), font='Any 12')
                continue

            # Validate output path
            output_ext = os.path.splitext(args['output_path'])[1].lower()
            if not any(output_ext == ext for ext in ('.csv', '.grd')):
                sg.PopupError(_('Output file extension must be *.csv (CSV table) or *.grd (ESRI ASCII grid)'),
                              title=_('Error'), font='Any 12')
                continue

            # Swap Conver/Cancel buttons
            swap_buttons(True)

            # Run converter process
            converter_process = mp.Process(target=run_cli, kwargs=args, daemon=False)
            converter_process.start()

            # Restore buttons state on process finished
            def restore_buttons():
                converter_process.join()
                swap_buttons(False)

            # Reset progress bar
            progress_bar.update(current_count=0)

            threading.Thread(target=restore_buttons, daemon=True).start()

    window.Close()



import multiprocessing as mp

import click
from click_option_group import optgroup

from .cli import run_cli
from .utils import GUI_ENABLED, install_i18n
from .version import version

if GUI_ENABLED:
    from .gui import run_gui


# Install localization
install_i18n()

# Add multiprocessing support on Windows for Pyinstaller's builds.
mp.freeze_support()


@click.version_option(version=version)
@click.command(help=_('QDC Converter.\n\nConverter of Garmin\'s QDC files into CSV or GRD.'))
@optgroup.group(_('Main parameters'), help=_('Key parameters of the converter'))
@optgroup.option('--qdc-folder-path', '-i', required=not(GUI_ENABLED),
                 type=click.Path(exists=True, resolve_path=True, file_okay=False, dir_okay=True),
                 help=_('Path to folder with QuickDraw Contours (QDC) inside.'))
@optgroup.option('--output-path', '-o', required=not(GUI_ENABLED),
                 type=click.Path(exists=False, resolve_path=True, file_okay=True, dir_okay=False),
                 help=_('Path to the result file (*.csv or *.grd).'))
@optgroup.option('--layer', '-l', required=not(GUI_ENABLED),
                 type=click.IntRange(0, 5), metavar='[0,1,2,3,4,5]',
                 help=_('Data layer (0 - Raw user data, 1 - Recommended).'))
@optgroup.group(_('Correction parameters'), help=_('Corrections'))
@optgroup.option('--x-correction', '-dx', type=click.FLOAT, default=0.0, help=_('Correction of X.'))
@optgroup.option('--y-correction', '-dy', type=click.FLOAT, default=0.0, help=_('Correction of Y.'))
@optgroup.option('--z-correction', '-dz', type=click.FLOAT, default=0.0, help=_('Correction of Z.'))
@optgroup.group(_('CSV parameters'), help=_('Parameters related to CSV'))
@optgroup.option('--csv-delimiter', '-csvd', type=click.STRING, default=',', help=_('CSV delimiter (default ",").'))
@optgroup.option('--csv-skip-headers', '-csvs', is_flag=True, help=_('Do not write header.'))
@optgroup.option('--csv-yxz', '-csvy', is_flag=True, help=_('Change column order from X,Y,Z to Y,X,Z.'))
@optgroup.group(_('Other parameters'), help=_('Other converter parameters'))
@optgroup.option('--singlethreaded', '-st', is_flag=True, help=_('Run converter in a single thread.'))
@optgroup.option('--validity-codes', '-vc', is_flag=True, help=_('Write validity code instead of depth.'))
@optgroup.option('--quite', '-q', is_flag=True, help=_('"Quite mode"'))
def main(qdc_folder_path, output_path, layer, validity_codes, quite, x_correction, y_correction, z_correction,
         csv_delimiter, csv_skip_headers, csv_yxz, singlethreaded):
    multithreaded = not singlethreaded
    if qdc_folder_path is None or output_path is None or layer is None:
        assert GUI_ENABLED
        return run_gui(qdc_folder_path, output_path, layer, validity_codes, quite, x_correction, y_correction,
                       z_correction, csv_delimiter, csv_skip_headers, csv_yxz, multithreaded)
    else:
        return run_cli(qdc_folder_path, output_path, layer, validity_codes, quite, x_correction, y_correction,
                       z_correction, csv_delimiter, csv_skip_headers, csv_yxz, multithreaded)



import csv
import multiprocessing as mp
import os
import struct
from functools import partial
from types import SimpleNamespace

import numpy as np
from tqdm import tqdm

from .cli import LAYER_PARAMETERS
from .utils import get_files_recursively, patch_tqdm, print_error

MULTIPROCESSING_BATCH = 64


def proc_init(shared_array):
    global SHARED_DEPTH_ARRAY
    SHARED_DEPTH_ARRAY = shared_array


def shared_depth_array(x_size, y_size):
    '''
    Returns shared depth array.
    '''
    np_array = np.frombuffer(SHARED_DEPTH_ARRAY.get_obj(),
                             dtype=np.dtype(SHARED_DEPTH_ARRAY.get_obj()._type_))
    np_array = np_array.reshape((x_size, y_size))
    return np_array


def calculate_grd_rows(j, validity_codes, x_size, y_size, z_correction):
    '''
    Multiprocessing GRD worker.
    '''
    result = []
    f_fix = lambda x: np.sign(x) * np.int16(np.abs(x))
    arr_depth = shared_depth_array(x_size, y_size)

    for i in range(x_size):
        if validity_codes:
            t_val = f_fix(arr_depth[i, j] / 4096)
            z = t_val * 10 + (arr_depth[i, j] - t_val * 4096) / 256
        else:
            z = arr_depth[i, j] / 100
        result.append(z + z_correction)

    return result


def calculate_csv_rows(j, validity_codes, x_size, y_size, x_orig, y_orig, layer):
    '''
    Multiprocessing CSV worker.
    '''
    result = []
    f_fix = lambda x: np.sign(x) * np.int16(np.abs(x))
    arr_depth = shared_depth_array(x_size, y_size)
    layer_parameters = SimpleNamespace(**LAYER_PARAMETERS[layer])

    for i in range(x_size):
        if arr_depth[i, j] > 0:  # Skip all 0 values
            if validity_codes:
                t_val = f_fix(arr_depth[i, j] / 4096)
                z = t_val * 10 + (arr_depth[i, j] - t_val * 4096) / 256
            else:
                z = arr_depth[i, j] / 100

            # Adding a_step / 2 to move point to the middle of the cell extent
            x = x_orig + layer_parameters.a_step / 2 + i * layer_parameters.a_step
            y = y_orig + layer_parameters.a_step / 2 + j * layer_parameters.a_step

            result.append((x, y, z))

    return result


def run_cli(qdc_folder_path, output_path, layer, validity_codes, quite, x_correction, y_correction, z_correction,
            csv_delimiter, csv_skip_headers, csv_yxz, message_queue=None, multithreaded=False):
    # Patch tqdm to duplicate messages up to the passed message queue.
    if message_queue:
        patch_tqdm(tqdm, message_queue)

    assert multithreaded

    try:
        # Some arguments validation
        output_path_ext = os.path.splitext(output_path)[-1]
        if output_path_ext.lower() not in ('.csv', '.grd'):
            raise ValueError(_('Output file extension must be *.csv (CSV table) or *.grd (ESRI ASCII grid)'))

        layer_parameters = SimpleNamespace(**LAYER_PARAMETERS[layer])
        qdc_files = get_files_recursively(qdc_folder_path, '.qdc')

        # Calculate boundaries
        x_min, y_min = 32000, 32000
        x_max, y_max = -32000, -32000
        for qdc_file in qdc_files:
            qdc_file_size = os.path.getsize(qdc_file)
            if qdc_file_size in (layer_parameters.f_size1, layer_parameters.f_size2,
                                 layer_parameters.f_size3, layer_parameters.f_size4):
                with open(qdc_file, 'rb') as f_qdc:
                    f_qdc.seek(164)
                    val = struct.unpack('<h', f_qdc.read(2))[0]
                    x_min = min(val, x_min)
                    x_max = max(val, x_max)

                    f_qdc.seek(160)
                    val = struct.unpack('<h', f_qdc.read(2))[0]
                    y_min = min(val, y_min)
                    y_max = max(val, y_max)

        if x_min == 32000 or y_min == 32000 \
           or x_max == -32000 or y_max == -32000:
            raise RuntimeError(_('No valid QDC files found!'))

        x_size = (x_max - x_min + 1) * layer_parameters.l_size
        y_size = (y_max - y_min + 1) * layer_parameters.l_size
        arr_depth = np.zeros((x_size, y_size), dtype=np.int16)

        # Calculate depth array
        for qdc_file in tqdm(qdc_files, desc=_('Calculating depth map'), disable=quite):
            qdc_file_size = os.path.getsize(qdc_file)
            if qdc_file_size in (layer_parameters.f_size1, layer_parameters.f_size2,
                                 layer_parameters.f_size3, layer_parameters.f_size4):
                with open(qdc_file, 'rb') as f_qdc:
                    f_qdc.seek(164)
                    val = struct.unpack('<h', f_qdc.read(2))[0]
                    x_orig = (val - x_min) * layer_parameters.l_size2

                    f_qdc.seek(160)
                    val = struct.unpack('<h', f_qdc.read(2))[0]
                    y_orig = (val - y_min) * layer_parameters.l_size2

                    if qdc_file_size == layer_parameters.f_size1:
                        i = layer_parameters.f_offset1
                    elif qdc_file_size == layer_parameters.f_size2:
                        i = layer_parameters.f_offset2
                    elif qdc_file_size == layer_parameters.f_size3:
                        i = layer_parameters.f_offset3
                    elif qdc_file_size == layer_parameters.f_size4:
                        i = layer_parameters.f_offset4

                    for yy in range(layer_parameters.n_sectors + 1):
                        for xx in range(layer_parameters.n_sectors + 1):
                            for y in range(32):
                                for x in range(32):
                                    x_abs = xx * 32 + x + x_orig
                                    y_abs = yy * 32 + y + y_orig
                                    f_qdc.seek(i + 1)
                                    val_code = struct.unpack('<h', f_qdc.read(2))[0]  # Read validity code

                                    if validity_codes:  # Write validity codes to array instead of depth
                                        arr_depth[x_abs, y_abs] = val_code
                                    else:
                                        if val_code != 0:
                                            f_qdc.seek(i - 1)
                                            val_depth = struct.unpack('<h', f_qdc.read(2))[0]  # Read depth in cm
                                            arr_depth[x_abs, y_abs] = val_depth
                                    i += 4

        x_orig = x_min * 90 / 2 ** 14
        y_orig = y_min * 90 / 2 ** 14

        # Create shared memory array that could be accessed from other
        # processes instead of pickling and passing it on each fork.
        global SHARED_DEPTH_ARRAY

        mp_array_typecode = np.ctypeslib.as_ctypes(arr_depth.dtype.type())._type_
        SHARED_DEPTH_ARRAY = mp.Array(typecode_or_type=mp_array_typecode,
                                      size_or_initializer=arr_depth.size)

        np_shared_depth_array = shared_depth_array(x_size, y_size)
        np.copyto(np_shared_depth_array, arr_depth)

        # Save depth array to *.csv or *.grd
        if output_path_ext.lower() == '.grd':
            # ESRI ASCII grid
            with open(output_path, 'w') as f_grd:
                f_grd.write(f'NCOLS {x_size}\n')
                f_grd.write(f'NROWS {y_size}\n')
                f_grd.write(f'XLLCORNER {x_orig}\n')
                f_grd.write(f'YLLCORNER {y_orig}\n')
                f_grd.write(f'CELLSIZE {layer_parameters.a_step}\n')
                f_grd.write('NODATA_VALUE 0\n')

                grd_rows_worker = partial(
                    calculate_grd_rows, validity_codes=validity_codes,
                    x_size=x_size, y_size=y_size, z_correction=z_correction
                )

                with mp.Pool(processes=mp.cpu_count(), initializer=proc_init, initargs=(SHARED_DEPTH_ARRAY,)) as pool:
                    y_rows = [*range(y_size - 1, -1, -1)]
                    pbar = iter(tqdm(y_rows, desc=_('Saving Esri ASCII raster'), disable=quite))
                    for rows in pool.imap(grd_rows_worker, y_rows, MULTIPROCESSING_BATCH):
                        next(pbar)
                        f_grd.write(' '.join(str(x) for x in rows) + '\n')

            # Write projection file
            output_path_prj = output_path[:-4] + '.prj'
            with open(output_path_prj, 'w') as f_prj:
                f_prj.write('GEOGCS["GCS_WGS_1984",DATUM["D_WGS_1984",'
                            'SPHEROID["WGS_1984",6378137.0,298.257223563]],'
                            'PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]]')

        elif output_path_ext.lower() == '.csv':
            # CSV table
            with open(output_path, 'w', newline='') as f_csv:
                writer = csv.writer(f_csv, delimiter=csv_delimiter)

                # Write header
                if not csv_skip_headers:
                    if csv_yxz:
                        if validity_codes:
                            writer.writerow(['Y', 'X', 'ValCode'])
                        else:
                            writer.writerow(['Y', 'X', 'Depth(m)'])
                    else:
                        if validity_codes:
                            writer.writerow(['X', 'Y', 'ValCode'])
                        else:
                            writer.writerow(['X', 'Y', 'Depth(m)'])

                # Write data
                csv_rows_worker = partial(
                    calculate_csv_rows, validity_codes=validity_codes,
                    x_size=x_size, y_size=y_size, x_orig=x_orig, y_orig=y_orig, layer=layer
                )

                with mp.Pool(processes=mp.cpu_count(), initializer=proc_init, initargs=(SHARED_DEPTH_ARRAY,)) as pool:
                    y_rows = [*range(y_size - 1, -1, -1)]
                    pbar = iter(tqdm(y_rows, desc=_('Saving CSV table'), disable=quite))
                    for rows in pool.imap(csv_rows_worker, y_rows, MULTIPROCESSING_BATCH):
                        next(pbar)
                        for x, y, z in rows:
                            if csv_yxz:
                                writer.writerow([y + y_correction, x + x_correction, z + z_correction])
                            else:
                                writer.writerow([x + x_correction, y + y_correction, z + z_correction])

    except Exception as e:
        print_error(f'{_("Error")}: {e}', message_queue)
        raise



import distutils.cmd
import os
import re
import sys

from setuptools import find_packages, setup


here = os.path.dirname(os.path.realpath(__file__))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(os.path.join(here, 'qdc_converter', 'version.py')) as f:
    match = re.search(r"^version\s*=\s*'(?P<version>.+?)'", f.read(), re.M)
    assert match
    VERSION = match.group('version')

deps = [
    'numpy>=1.19.0',
    'tqdm>=4.37.0',
    'click>=4.0.0',
    'click-option-group>=0.5.3',
]


class BuildStandaloneCommand(distutils.cmd.Command):
    """A custom command to build standalone app."""
    description = 'Build standalone app with PyInstaller'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import shutil
        import subprocess

        try:
            # Target filename
            dist_filename = 'QDCConverter'

            # Dist
            build_cmd = [
                'pyinstaller',
                '--clean',
                '--onefile',
                '--windowed',
                '-n', dist_filename,
            ]

            # Icon
            if sys.platform.startswith('win'):
                build_cmd += [
                    '--icon', 'qdc_converter/data/images/icon.ico',
                ]
            elif sys.platform.startswith('darwin'):
                build_cmd += [
                    '--icon', 'qdc_converter/data/images/icon.icns',
                ]

            # Add data
            build_cmd += [
                '--add-data', f'qdc_converter/data{os.pathsep}qdc_converter/data',
                'qdc-converter.py',
            ]

            print('Running command: %s' % ' '.join(build_cmd), file=sys.stderr)
            subprocess.check_call(build_cmd)
        finally:
            # Cleanup
            shutil.rmtree(os.path.join(here, 'build'), ignore_errors=True)
            try:
                os.remove(os.path.join(here, f'{dist_filename}.spec'))
            except FileNotFoundError:
                pass


setup(
    name='qdc-converter',
    description='Garmin QDC (Quickdraw Contours) Converter',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='qdc garmin converter',
    author='Andy Trofimov',
    author_email='interlark@gmail.com',
    url='http://github.com/interlark/qdc-converter',
    version=VERSION,
    python_requires='>=3.6',
    packages=find_packages(),
    include_package_data=True,
    install_requires=deps,
    extras_require={
        'gui': [
            'PySimpleGUI==4.60.0',
        ],
        'dev': [
            'pyinstaller>=5.0',
            'wheel>=0.36.2,<0.38',
            'pre-commit>=2.6',
            'pytest>=6.2,<8',
        ],
    },
    license='MIT',
    entry_points={'console_scripts': ['qdc-converter = qdc_converter:main']},
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Software Development',
        'Topic :: Terminals',
        'Topic :: Utilities',
    ],
    cmdclass={'build_standalone': BuildStandaloneCommand},
)



import csv
import os
import struct
from types import SimpleNamespace

import numpy as np
from tqdm import tqdm

from .utils import get_files_recursively, patch_tqdm, print_error


LAYER_PARAMETERS = {
    0: {
        'a_step': 90 / 2 ** 22,
        'n_sectors': 7,
        'f_size1': 372736,
        'f_offset1': 4097,
        'f_size2': 352256,
        'f_offset2': 4097,
        'f_size3': -1,
        'f_offset3': 0,
        'f_size4': -1,
        'f_offset4': 0,
        'l_size': 256,
        'l_size2': 256,
    },
    1: {
        'a_step': 90 / 2 ** 21,
        'n_sectors': 3,
        'f_size1': 372736,
        'f_offset1': 266241,
        'f_size2': 352256,
        'f_offset2': 266241,
        'f_size3': 110592,
        'f_offset3': 4097,
        'f_size4': 90112,
        'f_offset4': 4097,
        'l_size': 128,
        'l_size2': 128,
    },
    2: {
        'a_step': 90 / 2 ** 20,
        'n_sectors': 1,
        'f_size1': 372736,
        'f_offset1': 331777,
        'f_size2': 352256,
        'f_offset2': 331777,
        'f_size3': 110592,
        'f_offset3': 69633,
        'f_size4': 90112,
        'f_offset4': 69633,
        'l_size': 64,
        'l_size2': 64,
    },
    3: {
        'a_step': 90 / 2 ** 19,
        'n_sectors': 0,
        'f_size1': 372736,
        'f_offset1': 348161,
        'f_size2': 352256,
        'f_offset2': 348161,
        'f_size3': 110592,
        'f_offset3': 86017,
        'f_size4': 90112,
        'f_offset4': 86017,
        'l_size': 32,
        'l_size2': 32,
    },
    4: {
        'a_step': 90 / 2 ** 18,
        'n_sectors': 1,
        'f_size1': 372736,
        'f_offset1': 352257,
        'f_size2': -1,
        'f_offset2': 0,
        'f_size3': 110592,
        'f_offset3': 90113,
        'f_size4': -1,
        'f_offset4': 0,
        'l_size': 64,
        'l_size2': 16,
    },
    5: {
        'a_step': 90 / 2 ** 17,
        'n_sectors': 0,
        'f_size1': 372736,
        'f_offset1': 368641,
        'f_size2': -1,
        'f_offset2': 0,
        'f_size3': 110592,
        'f_offset3': 106497,
        'f_size4': -1,
        'f_offset4': 0,
        'l_size': 32,
        'l_size2': 8,
    },
}


def run_cli(qdc_folder_path, output_path, layer, validity_codes, quite, x_correction, y_correction, z_correction,
            csv_delimiter, csv_skip_headers, csv_yxz, message_queue=None, multithreaded=False):
    # Patch tqdm to duplicate messages up to the passed message queue.
    if message_queue:
        patch_tqdm(tqdm, message_queue)

    # Invoke multithreaded converter
    if multithreaded:
        from .cli_multithreaded import run_cli
        return run_cli(
            qdc_folder_path, output_path, layer, validity_codes, quite,
            x_correction, y_correction, z_correction,
            csv_delimiter, csv_skip_headers, csv_yxz, message_queue, multithreaded
        )

    try:
        # Some arguments validation
        output_path_ext = os.path.splitext(output_path)[-1]
        if output_path_ext.lower() not in ('.csv', '.grd'):
            raise ValueError(_('Output file extension must be *.csv (CSV table) or *.grd (ESRI ASCII grid)'))

        layer_parameters = SimpleNamespace(**LAYER_PARAMETERS[layer])
        qdc_files = get_files_recursively(qdc_folder_path, '.qdc')

        # Calculate boundaries
        x_min, y_min = 32000, 32000
        x_max, y_max = -32000, -32000
        for qdc_file in qdc_files:
            qdc_file_size = os.path.getsize(qdc_file)
            if qdc_file_size in (layer_parameters.f_size1, layer_parameters.f_size2,
                                 layer_parameters.f_size3, layer_parameters.f_size4):
                with open(qdc_file, 'rb') as f_qdc:
                    f_qdc.seek(164)
                    val = struct.unpack('<h', f_qdc.read(2))[0]
                    x_min = min(val, x_min)
                    x_max = max(val, x_max)

                    f_qdc.seek(160)
                    val = struct.unpack('<h', f_qdc.read(2))[0]
                    y_min = min(val, y_min)
                    y_max = max(val, y_max)

        if x_min == 32000 or y_min == 32000 \
           or x_max == -32000 or y_max == -32000:
            raise RuntimeError(_('No valid QDC files found!'))

        x_size = (x_max - x_min + 1) * layer_parameters.l_size
        y_size = (y_max - y_min + 1) * layer_parameters.l_size
        arr_depth = np.zeros((x_size, y_size), dtype=np.int16)

        # Calculate depth array
        for qdc_file in tqdm(qdc_files, desc=_('Calculating depth map'), disable=quite):
            qdc_file_size = os.path.getsize(qdc_file)
            if qdc_file_size in (layer_parameters.f_size1, layer_parameters.f_size2,
                                 layer_parameters.f_size3, layer_parameters.f_size4):
                with open(qdc_file, 'rb') as f_qdc:
                    f_qdc.seek(164)
                    val = struct.unpack('<h', f_qdc.read(2))[0]
                    x_orig = (val - x_min) * layer_parameters.l_size2

                    f_qdc.seek(160)
                    val = struct.unpack('<h', f_qdc.read(2))[0]
                    y_orig = (val - y_min) * layer_parameters.l_size2

                    if qdc_file_size == layer_parameters.f_size1:
                        i = layer_parameters.f_offset1
                    elif qdc_file_size == layer_parameters.f_size2:
                        i = layer_parameters.f_offset2
                    elif qdc_file_size == layer_parameters.f_size3:
                        i = layer_parameters.f_offset3
                    elif qdc_file_size == layer_parameters.f_size4:
                        i = layer_parameters.f_offset4

                    for yy in range(layer_parameters.n_sectors + 1):
                        for xx in range(layer_parameters.n_sectors + 1):
                            for y in range(32):
                                for x in range(32):
                                    x_abs = xx * 32 + x + x_orig
                                    y_abs = yy * 32 + y + y_orig
                                    f_qdc.seek(i + 1)
                                    val_code = struct.unpack('<h', f_qdc.read(2))[0]  # Read validity code

                                    if validity_codes:  # Write validity codes to array instead of depth
                                        arr_depth[x_abs, y_abs] = val_code
                                    else:
                                        if val_code != 0:
                                            f_qdc.seek(i - 1)
                                            val_depth = struct.unpack('<h', f_qdc.read(2))[0]  # Read depth in cm
                                            arr_depth[x_abs, y_abs] = val_depth
                                    i += 4

        x_orig = x_min * 90 / 2 ** 14
        y_orig = y_min * 90 / 2 ** 14

        f_fix = lambda x: np.sign(x) * np.int16(np.abs(x))

        # Save depth array to *.csv or *.grd
        if output_path_ext.lower() == '.grd':
            # ESRI ASCII grid
            with open(output_path, 'w') as f_grd:
                f_grd.write(f'NCOLS {x_size}\n')
                f_grd.write(f'NROWS {y_size}\n')
                f_grd.write(f'XLLCORNER {x_orig}\n')
                f_grd.write(f'YLLCORNER {y_orig}\n')
                f_grd.write(f'CELLSIZE {layer_parameters.a_step}\n')
                f_grd.write('NODATA_VALUE 0\n')

                for j in tqdm(range(y_size - 1, -1, -1), desc=_('Saving Esri ASCII raster')):
                    row_values = []
                    for i in range(x_size):
                        if validity_codes:
                            t_val = f_fix(arr_depth[i, j] / 4096)
                            z = t_val * 10 + (arr_depth[i, j] - t_val * 4096) / 256
                        else:
                            z = arr_depth[i, j] / 100
                        row_values.append(z + z_correction)
                    f_grd.write(' '.join(str(x) for x in row_values) + '\n')

            # Write projection file
            output_path_prj = output_path[:-4] + '.prj'
            with open(output_path_prj, 'w') as f_prj:
                f_prj.write('GEOGCS["GCS_WGS_1984",DATUM["D_WGS_1984",'
                            'SPHEROID["WGS_1984",6378137.0,298.257223563]],'
                            'PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]]')

        elif output_path_ext.lower() == '.csv':
            # CSV table
            with open(output_path, 'w', newline='') as f_csv:
                writer = csv.writer(f_csv, delimiter=csv_delimiter)

                # Write header
                if not csv_skip_headers:
                    if csv_yxz:
                        if validity_codes:
                            writer.writerow(['Y', 'X', 'ValCode'])
                        else:
                            writer.writerow(['Y', 'X', 'Depth(m)'])
                    else:
                        if validity_codes:
                            writer.writerow(['X', 'Y', 'ValCode'])
                        else:
                            writer.writerow(['X', 'Y', 'Depth(m)'])

                # Write data
                for j in tqdm(range(y_size - 1, -1, -1), desc=_('Saving CSV table'), disable=quite):
                    for i in range(x_size):
                        if arr_depth[i, j] > 0:  # Skip all 0 values
                            if validity_codes:
                                t_val = f_fix(arr_depth[i, j] / 4096)
                                z = t_val * 10 + (arr_depth[i, j] - t_val * 4096) / 256
                            else:
                                z = arr_depth[i, j] / 100

                            # Adding a_step / 2 to move point to the middle of the cell extent
                            x = x_orig + layer_parameters.a_step / 2 + i * layer_parameters.a_step
                            y = y_orig + layer_parameters.a_step / 2 + j * layer_parameters.a_step

                            if csv_yxz:
                                writer.writerow([y + y_correction, x + x_correction, z + z_correction])
                            else:
                                writer.writerow([x + x_correction, y + y_correction, z + z_correction])

    except Exception as e:
        print_error(f'{_("Error")}: {e}', message_queue)
        raise
